from typing import Callable, List


class Lambda:
    r'''
    Apply a user-defined lambda as a transform.

    .. code-block:: python

        transform = Lambda(lambda x: x + 1)
        transform(1)
        >>> 2

    Args:
        lambd (Callable): Lambda/function to be used for transform.

    .. automethod:: __call__
    '''
    def __init__(self, lambd: Callable):
        self.lambd = lambd

    def __call__(self, eeg: any) -> any:
        r'''
        Args:
            x (any): The input.

        Returns:
            any: The transformed output.
        '''
        return self.lambd(eeg)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"


class Compose:
    r'''
    Composes several transforms together. Consistent with :obj:`torchvision.transforms.Compose`'s behavior.

    .. code-block:: python

        transform = Composes([
            ToTensor(),
            Resize(size=(64, 64)),
            RandomNoise(p=0.1),
            RandomMask(p=0.1)
        ])
        transform(torch.randn(128, 9, 9)).shape
        >>> (128, 64, 64)

    :obj`Composes` supports transformers with different data dependencies. The above example combines multiple torch-based transformers, the following example shows a sequence of numpy-based transformer.

    .. code-block:: python

        transform = Composes([
            BandDifferentialEntropy(),
            MeanStdNormalize(),
            ToGrid(DEAP_CHANNEL_LOCATION_DICT)
        ])
        transform(np.random.randn(32, 128)).shape
        >>> (128, 9, 9)

    Args:
        transforms (list): The list of transforms to compose.

    .. automethod:: __call__
    '''
    def __init__(self, transforms: List[Callable]):
        self.transforms = transforms

    def __call__(self, eeg: any) -> any:
        r'''
        Args:
            x (any): The input.

        Returns:
            any: The transformed output.
        '''
        for t in self.transforms:
            eeg = t(eeg)
        return eeg

    def __repr__(self) -> str:
        format_string = self.__class__.__name__ + '('
        for t in self.transforms:
            format_string += '\n'
            format_string += '    {0}'.format(t)
        format_string += '\n)'
        return format_string


from typing import Dict, List, Union


class Select:
    r'''
    Select part of the value from the information dictionary.

    .. code-block:: python

        transform = Select(key='valence')
        transform({'valence': 4.5, 'arousal': 5.5, 'subject': 7})
        >>> 4.5

    :obj:`Select` allows multiple values to be selected and returned as a list. Suitable for multi-classification tasks or multi-task learning.

    .. code-block:: python

        transform = Select(key=['valence', 'arousal'])
        transform({'valence': 4.5, 'arousal': 5.5, 'subject': 7})
        >>> [4.5, 5.5]

    Args:
        key (str or list): The selected key can be a key string or a list of keys.

    .. automethod:: __call__
    '''
    def __init__(self, key: Union[str, List]):
        self.key = key
        self.select_list = isinstance(key, list) or isinstance(key, tuple)

    def __call__(self, label_dict: Dict) -> Union[int, float, List]:
        r'''
        Args:
            label_dict (dict): A dictionary describing the EEG signal samples, usually as the last return value for each sample in :obj:`Dataset`.
            
        Returns:
            str or list: Selected value or selected value list.
        '''
        assert isinstance(
            label_dict, dict
        ), f'The transform Select only accepts label dict as input, but obtain {type(label_dict)} as input.'
        if self.select_list:
            return [label_dict[k] for k in self.key]
        return label_dict[self.key]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"


class Binary:
    r'''
    Binarize the label according to a certain threshold. Labels larger than the threshold are set to 1, and labels smaller than the threshold are set to 0.
    
    .. code-block:: python

        transform = Binary(threshold=5.0)
        transform(4.5)
        >>> 0

    :obj:`Binary` allows simultaneous binarization using the same threshold for multiple labels.

    .. code-block:: python

        transform = Binary(threshold=5.0)
        transform([4.5, 5.5])
        >>> [0, 1]

    Args:
        threshold (float): Threshold used during binarization.

    .. automethod:: __call__
    '''
    def __init__(self, threshold: float):
        self.threshold = threshold

    def __call__(self, label: Union[int, float, List]) -> Union[int, List]:
        r'''
        Args:
            label (int, float, or list): The input label or list of labels.
            
        Returns:
            int, float, or list: The output label or list of labels after binarization.
        '''
        assert isinstance(
            label, (int, float, list)
        ), f'The transform Binary only accepts label list or item (int or float) as input, but obtain {type(label)} as input.'
        if isinstance(label, list):
            return [int(l >= self.threshold) for l in label]
        return int(label >= self.threshold)


class BinariesToCategory:
    r'''
    Convert multiple binary labels into one multiclass label. Multiclass labels represent permutations of binary labels. Commonly used to combine two binary classification tasks into a single quad classification task.
    
    .. code-block:: python

        transform = BinariesToCategory()
        transform([0, 0])
        >>> 0
        transform([0, 1])
        >>> 1
        transform([1, 0])
        >>> 2
        transform([1, 1])
        >>> 3
    
    .. automethod:: __call__
    '''
    def __call__(self, label_list: List) -> int:
        r'''
        Args:
            label_list (list): list of binary labels.
            
        Returns:
            int: The converted multiclass label.
        '''
        assert isinstance(
            label_list, list
        ), f'The transform BinariesToCategory only accepts label list as input, but obtain {type(label_list)} as input.'
        return sum([v * 2**i for i, v in enumerate(reversed(label_list))])



import torch
import numpy as np

from typing import List
from torch_geometric.data import Data


class ToG:
    r'''
    A transformation method for constructing a graph representation of EEG signals, the results of which are applied to the input of the :obj:`torch_geometric` model. In the graph, nodes correspond to electrodes, and edges correspond to associations between electrodes (eg, spatially adjacent or functionally connected)

    :obj:`TorchEEG` provides some common graph structures. Consider using the following adjacency matrices depending on the dataset (with different EEG acquisition systems):

    - datasets.constants.emotion_recognition.deap.DEAP_ADJACENCY_MATRIX
    - datasets.constants.emotion_recognition.dreamer.DREAMER_ADJACENCY_MATRIX
    - datasets.constants.emotion_recognition.seed.SEED_ADJACENCY_MATRIX
    - ...

    .. code-block:: python

        transform = ToG(adj=DEAP_ADJACENCY_MATRIX)
        transform(np.random.randn(32, 128)).shape
        >>> (32, 4)

    Args:
        adj (list): An adjacency matrix represented by a 2D array, each element in the adjacency matrix represents the electrode-to-electrode edge weight. Please keep the order of electrodes in the rows and columns of the adjacency matrix consistent with the EEG signal to be transformed.
        complete_graph (bool): Whether to build as a complete graph. If False, only construct edges between electrodes based on non-zero elements; if True, construct variables between all electrodes and set the weight of non-existing edges to 0. (defualt: :obj:`False`)
    
    .. automethod:: __call__
    '''
    def __init__(self, adj: List[List], complete_graph: bool = False):
        adj = torch.tensor(adj)
        if complete_graph:
            adj[adj == 0] = 1e-6
        self.adj = adj.to_sparse()

    def __call__(self, x: np.ndarray) -> Data:
        r'''
        Args:
            x (np.ndarray): The input EEG signals in shape of [number of electrodes, number of data points].

        Returns:
            torch_geometric.data.Data: The graph representation data types that torch_geometric can accept. Nodes correspond to electrodes, and edges are determined via the given adjacency matrix.
        '''
        data = Data(edge_index=self.adj._indices())
        data.x = torch.from_numpy(x).float()
        data.edge_weight = self.adj._values()

        return data


import os
import re
from copy import copy
from typing import List, Tuple, Union

import pandas as pd
from sklearn import model_selection
from torcheeg.datasets.module.base_dataset import BaseDataset


class KFoldTrialPerSubject:
    r'''
    A tool class for k-fold cross-validations, to divide the training set and the test set, commonly used to study model performance in the case of subject dependent experiments. Experiments were performed separately for each subject, where the data for all trials of the subject is divided into k subsets at the trial dimension, with one subset being retained as the test set and the remaining k-1 being used as training data. In most of the literature, K is chosen as 5 or 10 according to the size of the data set.

    .. code-block:: python

        cv = KFoldTrialPerSubject(n_splits=5, shuffle=True, split_path='./split')
        dataset = DEAPDataset(io_path=f'./deap',
                              root_path='./data_preprocessed_python',
                              online_transform=transforms.Compose([
                                  transforms.ToTensor(),
                                  transforms.To2d()
                              ]),
                              label_transform=transforms.Compose([
                                  transforms.Select(['valence', 'arousal']),
                                  transforms.Binary(5.0),
                                  transforms.BinariesToCategory()
                              ]))

        for train_dataset, test_dataset in cv.split(dataset):
            # The total number of experiments is the number subjects multiplied by K
            train_loader = DataLoader(train_dataset)
            test_loader = DataLoader(test_dataset)
            ...
    
    :obj:`KFoldTrialPerSubject` allows the user to specify the index of the subject of interest, when the user need to report the performance on each subject.

    .. code-block:: python

        cv = KFoldTrialPerSubject(n_splits=5, shuffle=True, split_path='./split')
        dataset = DEAPDataset(io_path=f'./deap',
                              root_path='./data_preprocessed_python',
                              online_transform=transforms.Compose([
                                  transforms.To2d(),
                                  transforms.ToTensor()
                              ]),
                              label_transform=transforms.Compose([
                                  transforms.Select(['valence', 'arousal']),
                                  transforms.Binary(5.0),
                                  transforms.BinariesToCategory()
                              ]))

        for train_dataset, test_dataset in cv.split(dataset, subject=1):
            # k-fold cross-validation for subject 1
            train_loader = DataLoader(train_dataset)
            test_loader = DataLoader(test_dataset)
            ...

    Args:
        n_splits (int): Number of folds. Must be at least 2. (default: :obj:`5`)
        shuffle (bool): Whether to shuffle the data before splitting into batches. Note that the samples within each split will not be shuffled. (default: :obj:`False`)
        random_state (int, optional): When shuffle is :obj:`True`, :obj:`random_state` affects the ordering of the indices, which controls the randomness of each fold. Otherwise, this parameter has no effect. (default: :obj:`None`)
        split_path (str): The path to data partition information. If the path exists, read the existing partition from the path. If the path does not exist, the current division method will be saved for next use. (default: :obj:`./split/k_fold_dataset`)
    '''
    def __init__(self,
                 n_splits: int = 5,
                 shuffle: bool = False,
                 random_state: Union[float, None] = None,
                 split_path: str = './split/k_fold_trial_per_subject'):
        self.n_splits = n_splits
        self.shuffle = shuffle
        self.random_state = random_state
        self.split_path = split_path

        self.k_fold = model_selection.KFold(n_splits=n_splits, shuffle=shuffle, random_state=random_state)

    def split_info_constructor(self, info: pd.DataFrame) -> None:
        subjects = list(set(info['subject']))
        for subject in subjects:
            subject_info = info[info['subject'] == subject]

            subject_train_infos = {}
            subject_test_infos = {}

            trail_ids = list(set(subject_info['trail_id']))
            for trail_id in trail_ids:
                trail_info = subject_info[subject_info['trail_id'] == trail_id]

                for i, (train_index, test_index) in enumerate(self.k_fold.split(trail_info)):
                    train_info = trail_info.iloc[train_index]
                    test_info = trail_info.iloc[test_index]

                    if not i in subject_train_infos:
                        subject_train_infos[i] = []

                    if not i in subject_test_infos:
                        subject_test_infos[i] = []

                    subject_train_infos[i].append(train_info)
                    subject_test_infos[i].append(test_info)

            for i in subject_train_infos.keys():
                subject_train_info = pd.concat(subject_train_infos[i], ignore_index=True)
                subject_test_info = pd.concat(subject_test_infos[i], ignore_index=True)
                subject_train_info.to_csv(os.path.join(self.split_path, f'train_subject_{subject}_fold_{i}.csv'),
                                          index=False)
                subject_test_info.to_csv(os.path.join(self.split_path, f'test_subject_{subject}_fold_{i}.csv'),
                                         index=False)

    @property
    def subjects(self) -> List:
        indice_files = list(os.listdir(self.split_path))

        def indice_file_to_subject(indice_file):
            return re.findall(r'subject_(\w*)_fold_(\d*).csv', indice_file)[0][0]

        return list(set(map(indice_file_to_subject, indice_files)))

    @property
    def fold_ids(self) -> List:
        indice_files = list(os.listdir(self.split_path))

        def indice_file_to_fold_id(indice_file):
            return int(re.findall(r'subject_(\w*)_fold_(\d*).csv', indice_file)[0][1])

        return list(set(map(indice_file_to_fold_id, indice_files)))

    def split(self, dataset: BaseDataset, subject: Union[int, None] = None) -> Tuple[BaseDataset, BaseDataset]:
        if not os.path.exists(self.split_path):
            os.makedirs(self.split_path)
            self.split_info_constructor(dataset.info)

        subjects = self.subjects
        fold_ids = self.fold_ids

        if not subject is None:
            assert subject in subjects, f'The subject should be in the subject list {subjects}.'

        for local_subject in subjects:
            if (not subject is None) and (local_subject != subject):
                continue

            for fold_id in fold_ids:
                train_info = pd.read_csv(
                    os.path.join(self.split_path, f'train_subject_{local_subject}_fold_{fold_id}.csv'))
                test_info = pd.read_csv(
                    os.path.join(self.split_path, f'test_subject_{local_subject}_fold_{fold_id}.csv'))

                train_dataset = copy(dataset)
                train_dataset.info = train_info

                test_dataset = copy(dataset)
                test_dataset.info = test_info

                yield train_dataset, test_dataset



from setuptools import setup, find_packages

__version__ = '1.0.2'
URL = 'https://github.com/tczhanzhi/torcheeg'

install_requires = [
    'tqdm',
    'numpy',
    'pandas',
    'scipy',
    'scikit-learn',
    'lmdb',
    'torch',
    'einops',
    'torch_geometric',
    'mne',
    'xmltodict'
]

test_requires = [
    'pytest',
    'pytest-cov',
]

example_requires = ['pytorch-lightning']

readme = open('README.rst').read()

setup(
    name='torcheeg',
    version='1.0.2',
    description=
    'TorchEEG is a library built on PyTorch for EEG signal analysis. TorchEEG aims to provide a plug-and-play EEG analysis tool, so that researchers can quickly reproduce EEG analysis work and start new EEG analysis research without paying attention to technical details unrelated to the research focus.',
    license='MIT',
    author='TorchEEG Team',
    author_email='tczhangzhi@gmail.com',
    keywords=['PyTorch', 'EEG'],
    url=URL,
    packages=find_packages(),
    long_description=readme,
    python_requires='>=3.8',
    extras_require={
        'example': example_requires,
        'test': test_requires
    },
    install_requires=install_requires)



from typing import Tuple

import torch
import torch.nn as nn


class MTCNN(nn.Module):
    r'''
    Multi-Task Convolutional Neural Network (MT-CNN). For more details, please refer to the following information.

    - Paper: Rudakov E, Laurent L, Cousin V, et al. Multi-Task CNN model for emotion recognition from EEG Brain maps[C]//2021 4th International Conference on Bio-Engineering for Smart Technologies (BioSMART). IEEE, 2021: 1-4.
    - URL: https://ieeexplore.ieee.org/abstract/document/9677807
    - Related Project: https://github.com/dolphin-in-a-coma/multi-task-cnn-eeg-emotion/

    Below is a recommended suite for use in emotion recognition tasks:

    .. code-block:: python
    
        DEAP_LOCATION_LIST = [['-', '-', 'AF3', 'FP1', '-', 'FP2', 'AF4', '-', '-'],
                              ['F7', '-', 'F3', '-', 'FZ', '-', 'F4', '-', 'F8'],
                              ['-', 'FC5', '-', 'FC1', '-', 'FC2', '-', 'FC6', '-'],
                              ['T7', '-', 'C3', '-', 'CZ', '-', 'C4', '-', 'T8'],
                              ['-', 'CP5', '-', 'CP1', '-', 'CP2', '-', 'CP6', '-'],
                              ['P7', '-', 'P3', '-', 'PZ', '-', 'P4', '-', 'P8'],
                              ['-', '-', '-', 'PO3', '-', 'PO4', '-', '-', '-'],
                              ['-', '-', '-', 'O1', 'OZ', 'O2', '-', '-', '-']]
        DEAP_CHANNEL_LOCATION_DICT = format_channel_location_dict(DEAP_CHANNEL_LIST, DEAP_LOCATION_LIST)

        dataset = DEAPDataset(io_path=f'./deap',
                    root_path='./data_preprocessed_python',
                    online_transform=transforms.Compose([
                        transforms.Concatenate([
                            transforms.BandDifferentialEntropy(),
                            transforms.BandPowerSpectralDensity()
                        ]),
                        transforms.ToGrid(DEAP_CHANNEL_LOCATION_DICT)
                    ]),
                    label_transform=transforms.Compose([
                        transforms.Select('valence'),
                        transforms.Binary(5.0),
                    ]))
        model = MTCNN(num_classes=2, in_channels=8, grid_size=(8, 9), dropout=0.2)

    Args:
        in_channels (int): The feature dimension of each electrode, i.e., :math:`N` in the paper. (defualt: :obj:`4`)
        grid_size (tuple): Spatial dimensions of grid-like EEG representation. (defualt: :obj:`(8, 9)`)
        num_classes (int): The number of classes to predict. (defualt: :obj:`2`)
        dropout (float): Probability of an element to be zeroed in the dropout layers. (defualt: :obj:`0.2`)
    '''
    def __init__(self,
                 in_channels: int = 8,
                 grid_size: Tuple[int, int] = (8, 9),
                 num_classes: int = 2,
                 dropout: float = 0.2):
        super(MTCNN, self).__init__()
        self.num_classes = num_classes
        self.in_channels = in_channels
        self.dropout = dropout
        self.grid_size = grid_size

        self.block1 = nn.Sequential(nn.Conv2d(in_channels, 64, kernel_size=5, padding=2, stride=1), nn.ReLU(),
                                    nn.BatchNorm2d(64), nn.Dropout2d(dropout))
        self.block2 = nn.Sequential(nn.ZeroPad2d((1, 2, 1, 2)), nn.Conv2d(64, 128, kernel_size=4, padding=0, stride=1),
                                    nn.ReLU(), nn.BatchNorm2d(128), nn.Dropout2d(dropout))
        self.block3 = nn.Sequential(nn.ZeroPad2d((1, 2, 1, 2)), nn.Conv2d(128, 256, kernel_size=4, padding=0, stride=1),
                                    nn.ReLU(), nn.BatchNorm2d(256), nn.Dropout2d(dropout))
        self.block4 = nn.Sequential(nn.Conv2d(256, 64, kernel_size=1, padding=0, stride=1), nn.ReLU(),
                                    nn.BatchNorm2d(64), nn.Dropout2d(dropout))
        self.lin1 = nn.Sequential(nn.Linear(grid_size[0] * grid_size[1] * 64, 512), nn.ReLU())
        self.lin1_bn = nn.Sequential(nn.BatchNorm1d(1),
                                  nn.Dropout(dropout))
        self.lin_v = nn.Linear(512, num_classes)
        self.lin_a = nn.Linear(512, num_classes)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        r'''
        Args:
            x (torch.Tensor): EEG signal representation, the ideal input shape is :obj:`[n, 8, 8, 9]`. Here, :obj:`n` corresponds to the batch size, :obj:`8` corresponds to :obj:`in_channels`, and :obj:`(8, 9)` corresponds to :obj:`grid_size`.

        Returns:
            torch.Tensor[number of sample, number of classes]: the predicted probability that the samples belong to the classes.
        '''
        x = self.block1(x)
        x = self.block2(x)
        x = self.block3(x)
        x = self.block4(x)

        x = x.view(x.size(0), -1)

        x = self.lin1(x)

        x = x.transpose(-1, -2)
        x = self.lin1_bn(x)
        x = x.transpose(-1, -2)

        return self.lin_v(x), self.lin_a(x)



import os
import random
import unittest
import shutil

import numpy as np

from torcheeg.io import EEGSignalIO


class TestEEGSignalIO(unittest.TestCase):
    def setUp(self):
        shutil.rmtree('./tmp_out/')
        os.mkdir('./tmp_out/')

    def test_init(self):
        io_cache_path = f'./tmp_out/{"".join(random.sample("zyxwvutsrqponmlkjihgfedcba", 20))}'
        io = EEGSignalIO(cache_path=io_cache_path)
        self.assertEqual(len(io), 0)

        self.assertTrue(os.path.exists(io_cache_path))
        io = EEGSignalIO(cache_path=io_cache_path)
        self.assertEqual(len(io), 0)

    def test_write_eeg(self):
        io_cache_path = f'./tmp_out/{"".join(random.sample("zyxwvutsrqponmlkjihgfedcba", 20))}'
        io = EEGSignalIO(cache_path=io_cache_path)

        eeg = np.random.randn(32, 128)
        io.write_eeg(eeg)

        eeg = np.random.randn(32, 128)
        io.write_eeg(eeg)

    def test_len(self):
        io_cache_path = f'./tmp_out/{"".join(random.sample("zyxwvutsrqponmlkjihgfedcba", 20))}'
        io = EEGSignalIO(cache_path=io_cache_path)

        eeg = np.random.randn(32, 128)
        io.write_eeg(eeg)

        eeg = np.random.randn(32, 128)
        io.write_eeg(eeg)

        self.assertEqual(len(io), 2)

    def test_read_eeg(self):
        io_cache_path = f'./tmp_out/{"".join(random.sample("zyxwvutsrqponmlkjihgfedcba", 20))}'
        io = EEGSignalIO(cache_path=io_cache_path)

        eeg = np.random.randn(32, 128)
        io_eeg_idx = io.write_eeg(eeg)

        io_eeg = io.read_eeg(io_eeg_idx)
        self.assertTrue(np.array_equal(eeg, io_eeg))


if __name__ == '__main__':
    unittest.main()


import unittest
import torch
import numpy as np
from torcheeg.transforms import ToTensor, Resize, RandomNoise, RandomMask


class TestTorchTransforms(unittest.TestCase):
    def test_to_tensor(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = ToTensor()(eeg)
        self.assertEqual(eeg.shape, tuple(transformed_eeg.shape))
        self.assertTrue(isinstance(transformed_eeg, torch.Tensor))

    def test_resize(self):
        eeg = torch.randn(128, 9, 9)
        transformed_eeg = Resize(size=(64, 64))(eeg)
        self.assertEqual(tuple(transformed_eeg.shape), (128, 64, 64))

    def test_random_noise(self):
        eeg = torch.randn(32, 128)
        transformed_eeg = RandomNoise()(eeg)
        self.assertEqual(tuple(transformed_eeg.shape), (32, 128))

    def test_random_mask(self):
        eeg = torch.randn(32, 128)
        transformed_eeg = RandomMask()(eeg)
        self.assertEqual(tuple(transformed_eeg.shape), (32, 128))


if __name__ == '__main__':
    unittest.main()


from typing import Callable, Union, Tuple, List

from ..base_dataset import BaseDataset
from ...functional.emotion_recognition.amigos import amigos_constructor
from ...constants.emotion_recognition.amigos import AMIGOS_CHANNEL_LOCATION_DICT, AMIGOS_ADJACENCY_MATRIX


class AMIGOSDataset(BaseDataset):
    r'''
    A dataset for Multimodal research of affect, personality traits and mood on Individuals and GrOupS (AMIGOS). This class generates training samples and test samples according to the given parameters, and caches the generated results in a unified input and output format (IO). The relevant information of the dataset is as follows:

    - Author: Miranda-Correa et al.
    - Year: 2018
    - Download URL: http://www.eecs.qmul.ac.uk/mmv/datasets/amigos/download.html
    - Reference: Miranda-Correa J A, Abadi M K, Sebe N, et al. Amigos: A dataset for affect, personality and mood research on individuals and groups[J]. IEEE Transactions on Affective Computing, 2018, 12(2): 479-493.
    - Stimulus: 16 short affective video extracts and 4 long affective video extracts from movies.
    - Signals: Electroencephalogram (14 channels at 128Hz), electrocardiogram (2 channels at 60Hz) and galvanic skin response (1 channel at 60Hz) of 40 subjects. For the first 16 trials, 40 subjects watched a set of short affective video extracts. For the last 4 trials, 37 of the participants of the previous experiment watched a set of long affective video extracts.
    - Rating: arousal (1-9), valence (1-9), dominance (1-9), liking (1-9), familiarity (1-9), neutral (0, 1), disgust (0, 1),happiness (0, 1), surprise (0, 1), anger (0, 1), fear (0, 1), and sadness (0, 1).

    In order to use this dataset, the download folder :obj:`data_preprocessed` is required, containing the following files:
    
    - Data_Preprocessed_P01.mat
    - Data_Preprocessed_P02.mat
    - Data_Preprocessed_P03.mat
    - ...
    - Data_Preprocessed_P40.mat

    An example dataset for CNN-based methods:

    .. code-block:: python

        dataset = AMIGOSDataset(io_path=f'./amigos',
                                root_path='./data_preprocessed',
                                offline_transform=transforms.Compose([
                                    transforms.BandDifferentialEntropy(),
                                    transforms.ToGrid(AMIGOS_CHANNEL_LOCATION_DICT)
                                ]),
                                online_transform=transforms.ToTensor(),
                                label_transform=transforms.Compose([
                                    transforms.Select('valence'),
                                    transforms.Binary(5.0),
                                ]))
        print(dataset[0])
        # EEG signal (torch.Tensor[128, 9, 9]),
        # coresponding baseline signal (torch.Tensor[128, 9, 9]),
        # label (int)

    Another example dataset for CNN-based methods:

    .. code-block:: python

        dataset = AMIGOSDataset(io_path=f'./amigos',
                              root_path='./data_preprocessed',
                              online_transform=transforms.Compose([
                                  transforms.To2d(),
                                  transforms.ToTensor()
                              ]),
                              label_transform=transforms.Compose([
                                  transforms.Select('valence'),
                                  transforms.Binary(5.0),
                              ]))
        print(dataset[0])
        # EEG signal (torch.Tensor[14, 128]),
        # coresponding baseline signal (torch.Tensor[14, 128]),
        # label (int)

    An example dataset for GNN-based methods:

    .. code-block:: python
    
        dataset = AMIGOSDataset(io_path=f'./amigos',
                              root_path='./data_preprocessed',
                              online_transform=transforms.Compose([
                                  transforms.ToG(AMIGOS_ADJACENCY_MATRIX)
                              ]),
                              label_transform=transforms.Compose([
                                  transforms.Select('valence'),
                                  transforms.Binary(5.0),
                              ]))
        print(dataset[0])
        # EEG signal (torch_geometric.data.Data),
        # coresponding baseline signal (torch_geometric.data.Data),
        # label (int)
        
    In particular, TorchEEG utilizes the producer-consumer model to allow multi-process data preprocessing. If your data preprocessing is time consuming, consider increasing :obj:`num_worker` for higher speedup.

    Args:
        root_path (str): Downloaded data files in matlab (unzipped data_preprocessed.zip) formats (default: :obj:`'./data_preprocessed'`)
        chunk_size (int): Number of data points included in each EEG chunk as training or test samples. (default: :obj:`128`)
        overlap (int): The number of overlapping data points between different chunks when dividing EEG chunks. (default: :obj:`0`)
        channel_num (int): Number of channels used, of which the first 14 channels are EEG signals. (default: :obj:`14`)
        trial_num (int): Number of trials used, of which the first 16 trials are conducted with short videos and the last 4 trials are conducted with long videos. If set to -1, all trials are used. (default: :obj:`16`)
        skipped_subjects (int): The participant ID to be removed because there are some invalid data in the preprocessed version. (default: :obj:`[9, 12, 21, 22, 23, 24, 33]`)
        baseline_num (int): Number of baseline signal chunks used. (default: :obj:`5`)
        baseline_chunk_size (int): Number of data points included in each baseline signal chunk. The baseline signal in the AMIGOS dataset has a total of 640 data points. (default: :obj:`128`)
        online_transform (Callable, optional): The transformation of the EEG signals and baseline EEG signals. The input is a :obj:`np.ndarray`, and the ouput is used as the first and second value of each element in the dataset. (default: :obj:`None`)
        offline_transform (Callable, optional): The usage is the same as :obj:`online_transform`, but executed before generating IO intermediate results. (default: :obj:`None`)
        label_transform (Callable, optional): The transformation of the label. The input is an information dictionary, and the ouput is used as the third value of each element in the dataset. (default: :obj:`None`)
        io_path (str): The path to generated unified data IO, cached as an intermediate result. (default: :obj:`./io/amigos`)
        num_worker (str): How many subprocesses to use for data processing. (default: :obj:`1`)
        verbose (bool): Whether to display logs during processing, such as progress bars, etc. (default: :obj:`True`)
    
    '''
    channel_location_dict = AMIGOS_CHANNEL_LOCATION_DICT
    adjacency_matrix = AMIGOS_ADJACENCY_MATRIX

    def __init__(self,
                 root_path: str = './data_preprocessed',
                 chunk_size: int = 128,
                 overlap: int = 0,
                 channel_num: int = 14,
                 trial_num: int = 16,
                 skipped_subjects: List[int] = [9, 12, 21, 22, 23, 24, 33],
                 baseline_num: int = 5,
                 baseline_chunk_size: int = 128,
                 online_transform: Union[None, Callable] = None,
                 offline_transform: Union[None, Callable] = None,
                 label_transform: Union[None, Callable] = None,
                 io_path: str = './io/amigos',
                 num_worker: int = 1,
                 verbose: bool = True):
        amigos_constructor(root_path=root_path,
                           chunk_size=chunk_size,
                           overlap=overlap,
                           channel_num=channel_num,
                           trial_num=trial_num,
                           skipped_subjects=skipped_subjects,
                           baseline_num=baseline_num,
                           baseline_chunk_size=baseline_chunk_size,
                           transform=offline_transform,
                           io_path=io_path,
                           num_worker=num_worker,
                           verbose=verbose)
        super().__init__(io_path)

        self.root_path = root_path
        self.chunk_size = chunk_size
        self.overlap = overlap
        self.channel_num = channel_num
        self.trial_num = trial_num
        self.skipped_subjects = skipped_subjects
        self.baseline_num = baseline_num
        self.baseline_chunk_size = baseline_chunk_size
        self.online_transform = online_transform
        self.offline_transform = offline_transform
        self.label_transform = label_transform
        self.io_path = io_path
        self.verbose = verbose

    def __getitem__(self, index: int) -> Tuple[any, any, int, int, int]:
        info = self.info.iloc[index].to_dict()

        eeg_index = str(info['clip_id'])
        eeg = self.eeg_io.read_eeg(eeg_index)

        if self.online_transform:
            eeg = self.online_transform(eeg)

        baseline_index = str(info['baseline_id'])
        baseline = self.eeg_io.read_eeg(baseline_index)

        if self.online_transform:
            baseline = self.online_transform(baseline)

        if self.label_transform:
            info = self.label_transform(info)

        if isinstance(info, list):
            return (eeg, baseline, *info)
        if isinstance(info, dict):
            return (eeg, baseline, *info.values())
        return eeg, baseline, info


import os
from torch.utils.data import Dataset
from abc import ABCMeta, abstractmethod

from torcheeg.io import EEGSignalIO, MetaInfoIO


class BaseDataset(Dataset, metaclass=ABCMeta):
    repr_indent = 4

    channel_location_dict = {}
    adjacency_matrix = []

    def __init__(self, io_path: str):
        if not self.exist(io_path):
            raise RuntimeError(
                'Database IO does not exist, please regenerate database IO.')
        self.io_path = io_path

        meta_info_io_path = os.path.join(self.io_path, 'info.csv')
        eeg_signal_io_path = os.path.join(self.io_path, 'eeg')

        info_io = MetaInfoIO(meta_info_io_path)
        self.eeg_io = EEGSignalIO(eeg_signal_io_path)

        self.info = info_io.read_all()

    def exist(self, io_path: str) -> bool:
        meta_info_io_path = os.path.join(io_path, 'info.csv')
        eeg_signal_io_path = eeg_signal_io_path = os.path.join(io_path, 'eeg')

        return os.path.exists(meta_info_io_path) and os.path.exists(
            eeg_signal_io_path)

    @abstractmethod
    def __getitem__(self, index: int) -> any:
        ...

    def __len__(self):
        return len(self.info)

    def __copy__(self) -> 'BaseDataset':
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        eeg_signal_io_path = os.path.join(self.io_path, 'eeg')

        result.eeg_io = EEGSignalIO(eeg_signal_io_path)
        return result

    def __repr__(self) -> str:
        head = "Dataset " + self.__class__.__name__
        body = ["Length: {}".format(self.__len__())]
        lines = [head] + [" " * self.repr_indent + line for line in body]
        return '\n'.join(lines)


import glob
import json
import os
from functools import partial
from multiprocessing import Manager, Pool, Process, Queue
from typing import Callable, List, Union

import mne
import xmltodict
from torcheeg.io import EEGSignalIO, MetaInfoIO
from tqdm import tqdm

MAX_QUEUE_SIZE = 1099511627776

def transform_producer(file_name: str, root_path: str, chunk_size: int, sampling_rate:int, overlap: int, channel_num: int,
                       baseline_num: int, baseline_chunk_size: int, trial_sample_num: int,
                       transform: Union[List[Callable], Callable, None], write_info_fn: Callable, queue: Queue):
    trial_dir = os.path.join(root_path, file_name)

    mne.set_log_level('CRITICAL')
    # record the common meta info for the trial
    label_file = os.path.join(trial_dir, 'session.xml')
    emodims = ['@feltArsl', '@feltCtrl', '@feltEmo', '@feltPred', '@feltVlnc', '@isStim']
    with open(label_file) as f:
        label_info = xmltodict.parse('\n'.join(f.readlines()))
    label_info = json.loads(json.dumps(label_info))['session']
    
    if not '@feltArsl' in label_info:
        # skip label_info['@isStim'] == '0' and other exception
        return

    trial_meta_info = {
        'subject_id': label_info['subject']['@id'],
        'trial_id': label_info['@mediaFile'],
        'duration': float(label_info['@cutLenSec'])
    }
    # feltArsl, feltCtrl, feltEmo, feltPred, feltVlnc, isStim
    trial_meta_info.update({k[1:]: int(label_info[k]) for k in emodims})

    write_pointer = 0

    # prepare transform
    if transform is None:
        transform = lambda x: x

    # extract signals
    sample_file = glob.glob(str(os.path.join(trial_dir, '*.bdf')))[0]

    raw = mne.io.read_raw_bdf(sample_file, preload=True, stim_channel='Status')
    events = mne.find_events(raw, stim_channel='Status')

    montage = mne.channels.make_standard_montage(kind='biosemi32')
    raw.set_montage(montage, on_missing='ignore')

    # pick channels
    raw.pick_channels(raw.ch_names[:channel_num])

    start_samp, end_samp = events[0][0] + 1, events[1][0] - 1

    # extract baseline signals
    trial_baseline_raw = raw.copy().crop(raw.times[0], raw.times[end_samp])
    trial_baseline_raw = trial_baseline_raw.resample(sampling_rate)

    trial_baseline_sample = trial_baseline_raw.to_data_frame().to_numpy()[:, 1:].swapaxes(
        1, 0)  # channel(32), timestep(30 * 128)
    trial_baseline_sample = trial_baseline_sample[:, :baseline_num * baseline_chunk_size]
    trial_baseline_sample = trial_baseline_sample.reshape(channel_num, baseline_num, baseline_chunk_size).mean(
        axis=1)  # channel(32), timestep(128)

    # put baseline signal into IO
    transformed_eeg = transform(trial_baseline_sample)
    trial_base_id = f'{file_name}_{write_pointer}'
    queue.put({'eeg': transformed_eeg, 'key': trial_base_id})
    write_pointer += 1
    trial_meta_info['baseline_id'] = trial_base_id

    # extract experimental signals
    trial_raw = raw.copy().crop(raw.times[start_samp], raw.times[end_samp])
    trial_raw = trial_raw.resample(sampling_rate)
    trial_samples = trial_raw.to_data_frame().to_numpy()[:, 1:].swapaxes(1, 0)

    # calculate moving step
    step = chunk_size - overlap

    start_at = 0
    end_at = start_at + chunk_size

    max_len = trial_samples.shape[1]
    if not (trial_sample_num <= 0):
        max_len = min(trial_sample_num * chunk_size, trial_samples.shape[1])

    while end_at <= max_len:
        clip_sample = trial_samples[:, start_at:end_at]
        transformed_eeg = transform(clip_sample)
        clip_id = f'{file_name}_{write_pointer}'
        queue.put({'eeg': transformed_eeg, 'key': clip_id})
        write_pointer += 1

        # record meta info for each signal
        record_info = {'start_at': start_at, 'end_at': end_at, 'clip_id': clip_id}
        record_info.update(trial_meta_info)
        write_info_fn(record_info)
        start_at = start_at + step
        end_at = start_at + chunk_size


def io_consumer(write_eeg_fn, queue):
    while True:
        item = queue.get()
        if not item is None:
            eeg = item['eeg']
            key = item['key']
            write_eeg_fn(eeg, key)
        else:
            break


def mahnob_constructor(root_path: str = './Sessions',
                       chunk_size: int = 128,
                       sampling_rate: int = 128,
                       overlap: int = 0,
                       channel_num: int = 32,
                       baseline_num: int = 30,
                       baseline_chunk_size: int = 128,
                       trial_sample_num: int = 30,
                       transform: Union[None, Callable] = None,
                       io_path: str = './io/mahnob',
                       num_worker: int = 1,
                       verbose: bool = True) -> None:
    # init IO
    if os.path.exists(io_path):
        if verbose:
            print(
                f'The target folder already exists, if you need to regenerate the database IO, please delete the path {io_path}.'
            )
        return
    os.makedirs(io_path)

    meta_info_io_path = os.path.join(io_path, 'info.csv')
    eeg_signal_io_path = os.path.join(io_path, 'eeg')

    info_io = MetaInfoIO(meta_info_io_path)
    eeg_io = EEGSignalIO(eeg_signal_io_path)

    # loop to access the dataset files
    file_list = os.listdir(root_path)

    if verbose:
        # show process bar
        pbar = tqdm(total=len(file_list))
        pbar.set_description("[MAHNOB]")

    manager = Manager()
    queue = manager.Queue(maxsize=MAX_QUEUE_SIZE)
    io_consumer_process = Process(target=io_consumer, args=(eeg_io.write_eeg, queue), daemon=True)
    io_consumer_process.start()

    partial_mp_fn = partial(transform_producer,
                            root_path=root_path,
                            chunk_size=chunk_size,
                            sampling_rate=sampling_rate,
                            overlap=overlap,
                            channel_num=channel_num,
                            baseline_num=baseline_num,
                            baseline_chunk_size=baseline_chunk_size,
                            trial_sample_num=trial_sample_num,
                            transform=transform,
                            write_info_fn=info_io.write_info,
                            queue=queue)

    for _ in Pool(num_worker).imap(partial_mp_fn, file_list):
        if verbose:
            pbar.update(1)

    if verbose:
        pbar.close()
        print('Please wait for the writing process to complete...')

    queue.put(None)

    io_consumer_process.join()
    io_consumer_process.close()



import unittest
import os
import random
import shutil

from torcheeg.model_selection import train_test_split_trial
from torcheeg.datasets import DEAPDataset


class TestTrainTestSplitTrial(unittest.TestCase):
    def setUp(self):
        shutil.rmtree('./tmp_out/')
        os.mkdir('./tmp_out/')

    def test_train_test_split_trial(self):
        io_path = f'./tmp_out/deap_{"".join(random.sample("zyxwvutsrqponmlkjihgfedcba", 20))}'
        root_path = './tmp_in/data_preprocessed_python'

        dataset = DEAPDataset(io_path=io_path, root_path=root_path)

        split_path = f'./tmp_out/split_{"".join(random.sample("zyxwvutsrqponmlkjihgfedcba", 20))}'

        train_dataset, test_dataset = train_test_split_trial(dataset=dataset, split_path=split_path)
        
        self.assertEqual(len(train_dataset), 61440)
        self.assertEqual(len(test_dataset), 15360)


if __name__ == '__main__':
    unittest.main()


import os
from typing import Tuple, Union

import lmdb
import numpy as np

MAX_LMDB_SIZE = 1099511627776


class EEGSignalIO:
    r'''
    A general-purpose, lightweight and efficient EEG signal IO APIs for converting various real-world EEG signal datasets into :obj:`np.ndarray` samples and storing them in the database. Here, we draw on the implementation ideas of industrial-grade application Caffe, and encapsulate a set of EEG signal reading and writing methods based on Lightning Memory-Mapped Database (LMDB), which not only unifies the differences of data types in different databases, but also accelerates the reading of data during training and testing.

    .. code-block:: python

        eeg_io = EEGSignalIO('YOUR_PATH')
        key = eeg_io.write_eeg(np.random.randn(32, 128))
        eeg = eeg_io.read_eeg(key)
        eeg.shape
        >>> (32, 128)
    
    Args:
        cache_path (str): Where the database is stored.
        cache_size (int): The maximum capacity of the database. (default: :obj:`1099511627776`)
    '''
    def __init__(self,
                 cache_path: str,
                 cache_size: int = MAX_LMDB_SIZE) -> None:
        self.cache_path = cache_path
        self.cache_size = cache_size

        if not os.path.exists(self.cache_path):
            os.makedirs(self.cache_path, exist_ok=True)

        self.db_eeg_size = None
        self.db_eeg_dtype = None

    @property
    def write_pointer(self):
        return len(self)

    def __len__(self):
        with lmdb.open(self.cache_path, self.cache_size, lock=False) as env:
            with env.begin() as transaction:
                # number of samples after removing db_eeg_size and db_eeg_dtype
                length = transaction.stat()['entries'] - 2
                return length if length > 0 else 0

    def set_eeg_format(self, eeg: np.ndarray) -> None:
        r'''
        Set the sample data shapes and types acceptable to the database based on the inserted EEG signals.

        Args:
            eeg (np.ndarray): The EEG signal sample to be written into the database.
        '''
        self.db_eeg_dtype = eeg.dtype
        self.db_eeg_size = eeg.shape

        with lmdb.open(self.cache_path, self.cache_size, lock=False) as env:
            with env.begin(write=True) as transaction:
                transaction.put('db_eeg_size'.encode(),
                                np.array(eeg.shape, dtype=np.uint32).tobytes())
                transaction.put('db_eeg_dtype'.encode(),
                                str(eeg.dtype).encode())

    def get_eeg_format(self) -> Tuple[np.ndarray, str]:
        r'''
        Get the sample data shapes and types acceptable to the database.

        Returns:
            tuple (np.ndarray, str): the sample data shapes and types acceptable to the database.
        '''
        with lmdb.open(self.cache_path, self.cache_size, lock=False) as env:
            with env.begin() as transaction:
                db_eeg_size = transaction.get('db_eeg_size'.encode())
                db_eeg_dtype = transaction.get('db_eeg_dtype'.encode())
            if (db_eeg_size is None) or (db_eeg_dtype is None):
                return (None, None)
            return tuple(np.frombuffer(
                db_eeg_size, dtype=np.uint32).tolist()), db_eeg_dtype.decode()

    def write_eeg(self, eeg: np.ndarray, key: Union[str, None] = None) -> str:
        r'''
        Write EEG signal to database.

        Args:
            eeg (np.ndarray): EEG signal samples to be written into the database.
            key (str, optional): The key of the EEG signal to be inserted, if not specified, it will be an auto-incrementing integer.

        Returns:
            int: The index of written EEG signals in the database.
        '''
        if (self.db_eeg_size is None) or (self.db_eeg_dtype is None):
            self.db_eeg_size, self.db_eeg_dtype = self.get_eeg_format()
            if (self.db_eeg_size is None) or (self.db_eeg_dtype is None):
                self.set_eeg_format(eeg)
        elif self.db_eeg_size != eeg.shape:
            raise RuntimeError(
                f'Inserting eeg sample failed. The dimensions of the eeg array in the database, {self.db_eeg_size}, do not match the sample to be inserted, {eeg.shape}.'
            )
        elif self.db_eeg_dtype != eeg.dtype:
            raise RuntimeError(
                f'Inserting eeg sample failed. The dtype of the eeg array in the database, {self.db_eeg_dtype}, do not match the sample to be inserted, {eeg.dtype}.'
            )

        if key is None:
            key = str(self.write_pointer)

        with lmdb.open(self.cache_path, self.cache_size, lock=False) as env:
            with env.begin(write=True) as transaction:
                transaction.put(key.encode(), eeg.tobytes())
            return key

    def read_eeg(self, key: str) -> np.ndarray:
        r'''
        Query the corresponding EEG signal in the database according to the index.

        Args:
            key (str): The index of the EEG signal to be queried.
            
        Returns:
            np.ndarray: The EEG signal sample.
        '''
        if (self.db_eeg_size is None) or (self.db_eeg_dtype is None):
            self.db_eeg_size, self.db_eeg_dtype = self.get_eeg_format()

        with lmdb.open(self.cache_path, self.cache_size, lock=False) as env:
            with env.begin() as transaction:
                eeg = transaction.get(key.encode())
            if eeg is None:
                raise RuntimeError(f'Unable to index the EEG signal sample with key {key}!')
            return np.frombuffer(eeg, dtype=self.db_eeg_dtype).reshape(
                self.db_eeg_size)


from typing import Tuple

import torch
import torch.nn as nn


class InceptionConv2d(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, bias: bool = True):
        super().__init__()
        self.conv5x5 = nn.Conv2d(in_channels, out_channels, kernel_size=5, stride=1, padding=2, bias=bias)
        self.conv3x3 = nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=1, padding=1, bias=bias)
        self.conv1x1 = nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=1, padding=0, bias=bias)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.conv5x5(x) + self.conv3x3(x) + self.conv1x1(x)


class SeparableConv2d(nn.Module):
    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: int = 3,
                 stride: int = 1,
                 padding: int = 1,
                 bias: bool = True):
        super().__init__()
        self.depth = nn.Conv2d(in_channels,
                               in_channels,
                               kernel_size=kernel_size,
                               stride=stride,
                               padding=padding,
                               groups=in_channels,
                               bias=bias)
        self.point = nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=stride, padding=0, bias=bias)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.depth(x)
        x = self.point(x)
        return x


class STNet(nn.Module):
    r'''
    Spatio-temporal Network (STNet). For more details, please refer to the following information.

    - Paper: Zhang Z, Zhong S, Liu Y. GANSER: A Self-supervised Data Augmentation Framework for EEG-based Emotion Recognition[J]. IEEE Transactions on Affective Computing, 2022.
    - URL: https://ieeexplore.ieee.org/abstract/document/9763358/
    - Related Project: https://github.com/tczhangzhi/GANSER-PyTorch

    Below is a recommended suite for use in emotion recognition tasks:

    .. code-block:: python

        dataset = DEAPDataset(io_path=f'./deap',
                    root_path='./data_preprocessed_python',
                    offline_transform=transforms.Compose([
                        transforms.ToGrid(DEAP_CHANNEL_LOCATION_DICT)
                    ]),
                    online_transform=transforms.ToTensor(),
                    label_transform=transforms.Compose([
                        transforms.Select('valence'),
                        transforms.Binary(5.0),
                    ]))
        model = STNet(num_classes=2, in_channels=4, grid_size=(9, 9), dropout=0.2)

    Args:
        in_channels (int): The dimension of each electrode. (defualt: :obj:`128`)
        grid_size (tuple): Spatial dimensions of grid-like EEG representation. (defualt: :obj:`(9, 9)`)
        num_classes (int): The number of classes to predict. (defualt: :obj:`2`)
        dropout (float): Probability of an element to be zeroed in the dropout layers. (defualt: :obj:`0.2`)
    '''
    def __init__(self,
                 in_channels: int = 128,
                 grid_size: Tuple[int, int] = (9, 9),
                 num_classes: int = 2,
                 dropout: float = 0.2):
        super(STNet, self).__init__()
        self.num_classes = num_classes
        self.in_channels = in_channels
        self.dropout = dropout
        self.grid_size = grid_size

        self.layer1 = nn.Conv2d(in_channels, 256, kernel_size=3, stride=1, padding=1, bias=True)
        self.layer2 = nn.Conv2d(256, 128, kernel_size=5, stride=1, padding=2, bias=True)
        self.layer3 = nn.Conv2d(128, 64, kernel_size=5, stride=1, padding=2, bias=True)
        self.layer4 = SeparableConv2d(64, 32, kernel_size=5, stride=1, padding=2, bias=True)
        self.layer5 = InceptionConv2d(32, 16)

        self.drop_selu = nn.Sequential(nn.Dropout(p=dropout), nn.SELU())

        self.lin1 = nn.Linear(grid_size[0] * grid_size[1] * 16, 1024, bias=True)
        self.lin2 = nn.Linear(1024, num_classes, bias=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        r'''
        Args:
            x (torch.Tensor): EEG signal representation, the ideal input shape is :obj:`[n, 128, 9, 9]`. Here, :obj:`n` corresponds to the batch size, :obj:`128` corresponds to :obj:`in_channels`, and :obj:`(9, 9)` corresponds to :obj:`grid_size`.

        Returns:
            torch.Tensor[number of sample, number of classes]: the predicted probability that the samples belong to the classes.
        '''
        x = self.layer1(x)
        x = self.drop_selu(x)
        x = self.layer2(x)
        x = self.drop_selu(x)
        x = self.layer3(x)
        x = self.drop_selu(x)
        x = self.layer4(x)
        x = self.drop_selu(x)
        x = self.layer5(x)
        x = self.drop_selu(x)
        x = x.view(x.size(0), -1)
        x = self.lin1(x)
        x = self.drop_selu(x)
        x = self.lin2(x)
        return x
